#!/usr/bin/env python3
import itertools
from typing import Iterator, Optional

__all__ = ["peek"]

import logging

logger = logging.getLogger(__name__)


def peek(generator: Iterator) -> Optional[itertools.chain]:
    try:
        return itertools.chain((next(generator),), generator)
    except StopIteration:
        return None


if __name__ == "__main__":
    logger.info(peek(iter(range(0))))
    logger.info(peek(iter(range(1))))



#!/usr/bin/env python3

__doc__ = r"""

           Created on 31/07/2023
           """

__author__ = "Christian Heider Lindbjerg"

from pathlib import Path

with open(Path(__file__).parent / "README.md") as this_init_file:
    __doc__ += this_init_file.read()

from .pip_parsing import *
from .reloading import *
from .editable import *
from .versioning import *
from .namespaces import *
from .satisfaction import *



#!/usr/bin/env python3


__project__ = "Warg"

__author__ = "Christian Heider Lindbjerg"
__version__ = "1.4.9"
__doc__ = r"""
Created on 27/04/2019

@author: cnheider

"""

from pathlib import Path

import logging

logger = logging.getLogger(__name__)

with open(Path(__file__).parent / "README.md") as this_init_file:
    __doc__ += this_init_file.read()
# del Path

# with open(Path(__file__).parent.parent / "README.md", "r") as this_init_file:
#    __doc__ += this_init_file.read()

# __all__ = ["PROJECT_APP_PATH", "PROJECT_NAME", "PROJECT_VERSION", "get_version"] # let everything be accessible
# from base warg import

try:
    # from .ode import * # Silly thing
    from .data_structures import *
    from .arguments import *
    from .gdkc import *
    from .mixins import *
    from .decorators import *
    from .metas import *
    from .bases import *
    from .typing_extension import *
    from .context_wrapper import *
    from .boolean_tests import *
    from .map_itertools import *
    from .ast_ops import *
    from .functions import *
    from .os_utilities import *
    from .generators import *
    from .text import *
    from .math_utilities import *
    from .business import *
    from .datetimes import *
    from .debug import *
    from .exceptions import *
    from .manipulation import *
    from .replication import *
    from .strings import *
    from .contexts import *
    from .config_shell import *
    from .colors import *
    from .packages import *
    from .iteration import *
    from .logging_utilities import *
    from .modules import *
except ImportError as ix:
    this_package_name = Path(__file__).parent.name
    this_package_reqs = Path(__file__).parent.parent / f"requirements.txt"
    if this_package_reqs.exists():
        logger.info(
            f"Make sure requirements is installed for {this_package_name}, see {this_package_reqs}"
        )  # TODO: PARSE WHAT is missing and print
    raise ix

PROJECT_NAME = clean_string(__project__)
PROJECT_VERSION = __version__
PROJECT_YEAR = 2018
PROJECT_AUTHOR = clean_string(__author__)
PROJECT_ORGANISATION = clean_string("Pything")

__url__ = f"https://github.com/{PROJECT_ORGANISATION}/{PROJECT_NAME}"

# from apppath import AppPath # CAREFUL CIRCULAR DEPENDENCY WARNING!
# PROJECT_APP_PATH = AppPath(app_name=PROJECT_NAME, app_author=PROJECT_AUTHOR) # NOT USED!

import_issue_found = False
try:
    from importlib.resources import files
    from importlib.metadata import PackageNotFoundError
except:
    try:
        from importlib_metadata import PackageNotFoundError
        from importlib_resources import files
    except:
        import_issue_found = True

if import_issue_found:
    PACKAGE_DATA_PATH = Path(__file__).parent / "data"
    DEVELOP = False
else:
    PACKAGE_DATA_PATH = files(PROJECT_NAME) / "data"

    try:
        DEVELOP = package_is_editable(PROJECT_NAME)
    except PackageNotFoundError as e:
        DEVELOP = True

__version__ = get_version(__version__, append_time=DEVELOP)
__version_info__ = tuple(int(segment) for segment in __version__.split("."))

if __name__ == "__main__":
    logger.info(__version__)



#!/usr/bin/env python3

__author__ = "Christian Heider Lindbjerg"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

__all__ = ["SingletonBase", "SingletonMeta", "key_singleton", "singleton"]

import functools
import logging
from functools import wraps
from typing import Any, Callable, MutableMapping, Sequence

logger = logging.getLogger(__name__)


class SingletonBase:
    """
    A base class for creating singleton class where all subtypes(Derivations) should also return the first
    and only
    instantiation of a particular singleton base type, if this property is not wanted consider using the
    SingletonMeta class instead."""

    instance = None

    def __new__(cls, *args: Sequence[Any], **kwargs: MutableMapping[str, Any]):
        if cls.instance is None:
            cls.instance = super().__new__(cls, *args, **kwargs)

        return cls.instance


class SingletonMeta(type):
    """
    Conversely the SingletonBase, this base meta class is used for creating singleton class where all
    subtypes(
    Deriavations) should only
    return
    singleton instantiations of a particular singleton type independantly of subtyping and super-types,
    if this property is not
    wanted
    consider using
    the
    SingletonBase class instead."""

    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        cls.instance = None

    def __call__(cls, *args: Sequence[Any], **kwargs: MutableMapping[str, Any]):
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kwargs)

        return cls.instance


def singleton(cls):
    """Use class as singleton."""

    cls.__new_original__ = cls.__new__

    @functools.wraps(cls.__new__)
    def singleton_new(cls_, *args: Sequence[Any], **kwargs: MutableMapping[str, Any]):
        """

        :param cls_:
        :param args:
        :param kwargs:
        :return:
        """
        it = cls_.__dict__.get("__it__")
        if it is not None:
            return it

        cls_.__it__ = it = cls_.__new_original__(cls_, *args, **kwargs)
        it.__init_original__(*args, **kwargs)
        return it

    cls.__new__ = singleton_new
    cls.__init_original__ = cls.__init__
    cls.__init__ = object.__init__

    return cls


def key_singleton(cache_key: Any) -> Callable:
    """
    TODO: finish
    """

    def inner_fn(fn: Callable) -> Any:
        """description"""

        @wraps(fn)
        def wrapper(self, *args: Sequence[Any], **kwargs: MutableMapping[str, Any]):
            """description"""
            instance = getattr(self, cache_key)
            if instance is not None:
                return instance

            instance = fn(self, *args, **kwargs)
            setattr(self, cache_key, instance)
            return instance

        return wrapper

    return inner_fn


if __name__ == "__main__":

    class SingletonBaseClass(SingletonBase):
        pass

    class S1(SingletonBaseClass):
        pass

    class SingletonBaseMeta(metaclass=SingletonMeta):
        pass

    class S2(SingletonBaseMeta):
        pass

    # expected
    logger.info(SingletonBaseClass())  # same
    logger.info(SingletonBaseClass())  # same
    logger.info(S1())  # same

    # expected
    logger.info(SingletonBaseMeta())  # same
    logger.info(SingletonBaseMeta())  # same
    logger.info(S2())  # different



from warg import leaf_apply, leaf_type_apply


import logging

logger = logging.getLogger(__name__)


def test_leaf_apply():
    asd = [1, 2, 3, 4, 5, 6, 7, 8, 9, [2, 3], [[323, 525], [323, 525], [323]]]
    logger.info(leaf_apply(asd, lambda a: a + 1))


def test_leaf_type_apply():
    asd = [(1, 2), [(3, 4), (3, 4)], [(323, 525)], [[(323, 525), (323, 39)]]]
    logger.info(leaf_type_apply(asd, lambda a: ((a[0] + a[1]), 1), tuple))


def test_reversed_ok():
    a = range(10)
    assert list(reversed(a)) == list(a)[::-1]


def test_reversed_zip():
    a = range(10)
    a = reversed(a)
    b = iter(a)
    next(b)
    c = zip(a, b)
    assert True  # WTF below?
    # assert len(list(c)) == 9  # ?????? BUG


def test_forward_zip():
    a = range(10)
    b = iter(a)
    next(b)
    c = zip(a, b)
    assert len(list(c)) == 9  # WORKS



import importlib
import logging
import pkgutil
from types import ModuleType
from typing import Any, Dict

logger = logging.getLogger(__name__)
__all__ = ["import_submodules", "import_submodule_alls"]


def import_submodules(package: Any, recursive: bool = True) -> Dict[str, ModuleType]:
    """Import all submodules of a module, recursively, including subpackages

    :param package: package (name or actual module)
    :param recursive:
    """

    if isinstance(package, str):
        if package == "__main__":
            assert NotImplementedError
        else:
            package = importlib.import_module(package)

    results = {}
    for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = f"{str(package.__name__)}.{str(name)}"
        results[full_name] = importlib.import_module(full_name)

        if recursive and is_pkg:
            results.update(import_submodules(full_name))

    return results


def import_submodule_alls(package: Any) -> Dict[str, ModuleType]:
    """Import all submodules of a module, recursively, including subpackages

    :param package: package (name or actual module)
    :param recursive:
    """

    if isinstance(package, str):
        if package == "__main__":
            assert NotImplementedError
        else:
            package = importlib.import_module(package)

    results = {}
    for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = f"{str(package.__name__)}.{str(name)}"
        sub_module = importlib.import_module(full_name)
        if hasattr(sub_module, "__all__"):
            alls = sub_module.__all__
        else:
            alls = dir(sub_module)
        for a in alls:
            results[a] = getattr(sub_module, a)

    return results



#!/usr/bin/env python3
from typing import Any, Callable, Iterable, Mapping, Tuple

__author__ = "Christian Heider Lindbjerg"
__doc__ = r"""

           Created on 11/11/2019
           """

__all__ = ["yield_and_map", "inner_map", "kw_map"]

import logging

logger = logging.getLogger(__name__)


def yield_and_map(iterable: Iterable[Any], level: int = 0, func: Callable = print) -> Any:
    """

    :param iterable:
    :type iterable:
    :param level:
    :type level:
    :param func:
    :type func:"""
    if level == 0:
        for a in iterable:
            func(a)
            yield a
    elif level == 1:
        for a in iterable:
            for b in a:
                func(b)
                yield b
    elif level == 2:
        for a in iterable:
            for b in a:
                for c in b:
                    func(c)
                    yield c


def inner_map(func: Callable, iterable: Iterable[Any], aggregate_yield: bool = True) -> Any:
    """

    :param func:
    :type func:
    :param iterable:
    :type iterable:
    :param aggregate_yield:
    :type aggregate_yield:"""
    if aggregate_yield:
        for a in iterable:
            yield [func(b) for b in a]
    else:
        for a in iterable:
            for b in a:
                yield func(b)


def kw_map(func: Callable, kw: str, iterable: Iterable[Any]) -> Any:
    """

    :param func:
    :type func:
    :param kw:
    :type kw:
    :param iterable:
    :type iterable:"""
    for a in iterable:
        yield func(**{kw: a})


def select_key(tuple_iterator: Iterable[Any], *a) -> Tuple[Any, Any]:
    """
    Yield keys from mapping if in a

    TODO: Why return the key itself?

    :param tuple_iterator:
    :type tuple_iterator:
    :param a:
    :type a:
    :return:
    :rtype:
    """
    for k, _ in tuple_iterator:
        if k in a:
            yield k, _


def select_dict(mapping: Mapping[Any, Any], *a) -> Mapping[Any, Any]:
    """
    Select keys from mapping if in a

    Args:
      mapping:
      *a:

    Returns:

    """
    return {k: v for k, v in select_key(mapping.items(), *a)}


if __name__ == "__main__":

    def uahsd() -> None:
        agfas = (2, 3)
        # TODO

        sadd = {a: b for a, b in zip("abcdef", range(6))}
        logger.info(select_dict(sadd, "a", "d"))

    uahsd()



#!/usr/bin/env python3
import logging
from typing import Callable, Dict, Hashable, Iterable, Mapping, MutableMapping
from collections import defaultdict


logger = logging.getLogger(__name__)
__all__ = [
    "invert_mapping",
    "invert_dict",
    "AppendingDict",
    "pivot_dict_object",
    "pivot_dict",
    "to_dict",
    "nested_dict",
]


def append_to_dict(d: Dict, key, value) -> Dict:
    """

    :param d:
    :type d:
    :param key:
    :type key:
    :param value:
    :type value:
    :return:
    :rtype:
    """
    d.setdefault(key, []).append(value)
    return d


class AppendingDict(Dict):  # appending_dict = collections.defaultdict(list)
    def __setitem__(self, key, value):
        # self.setdefault(key, []).append(value)
        # append_to_dict(self, key, value)
        if key in self:
            self[key].append(value)
        else:
            super().__setitem__(key, [value])


def recurse_mapping(a: Mapping, call: Callable = print) -> None:
    for k, v in a.items():
        if isinstance(v, Mapping):
            recurse_mapping(v)
        call(v)


def invert_mapping(m: Mapping) -> Mapping:  # TODO: TEST THIS; MAY CONTAINS BUGS!
    """
    Invert a mapping

    if a mapping does not have duplicate hashable values, then this is the same as invert_dict, otherwise values in
    new_m are tuples of keys with duplicate values
    :return: :rtype:
    """

    if isinstance(m, MutableMapping):
        new_m = type(m)()
    else:
        new_m = {}

    for k, v in m.items():
        if not isinstance(v, Hashable):
            raise TypeError(f"values must be hashable, was {type(v), v}, for key {k}")
        if v in new_m:
            if isinstance(new_m[v], Iterable) and (not isinstance(new_m[v], str)):
                new_m[v] = (*new_m[v], k)
            else:
                new_m[v] = [new_m[v], k]
        else:
            new_m[v] = k
    return new_m


def invert_dict(d: Mapping) -> Dict:  # TODO: HANDLE DUPLICATE KEYS; CONVERT TO TUPLES
    """
    Invert a dict

    :param d:
    :type d:
    :return:
    :rtype:
    """
    return {v: k for k, v in d.items()}


def pivot_dict(d: Dict, key) -> Dict:  # TODO: HANDLE DUPLICATE KEYS; CONVERT TO TUPLES
    """
    pivot_key -> pivot_value

    :param d:
    :param key:
    :return:
    :rtype:
    """
    return {v[key]: k for k, v in d.items()}


def pivot_dict_object(d: Dict, key) -> Dict:  # TODO: HANDLE DUPLICATE KEYS; CONVERT TO TUPLES
    """
    pivot_key -> pivot_value for object attributes

    :param d:
    :param key:
    :return:
    :rtype:
    """
    return {getattr(v, key): k for k, v in d.items()}


def to_dict(m: Mapping) -> dict:
    o = {}
    for k, v in m.items():
        if isinstance(v, Mapping):
            o[k] = to_dict(v)
        elif isinstance(v, str):
            o[k] = v
        elif isinstance(v, Iterable):
            o[k] = [to_dict(v_) if isinstance(v_, Mapping) else v_ for v_ in v]
        else:
            o[k] = v
    return o


nested_dict = lambda: defaultdict(nested_dict)

if __name__ == "__main__":
    logger.info(invert_mapping({"a": 1, "b": 2}))

    logger.info(invert_mapping({"a": 2, "b": 2, "c": 3, "d": 4}))
    logger.info(invert_dict({"a": 2, "b": 2, "c": 3, "d": 4}))

    def uhasdu():
        from warg.data_structures.named_ordered_dictionary import NOD

        a = NOD({"b": NOD(c=1)}, d="usahfo7uyhaouw", f=[NOD(p="m")])

        logger.info(a)
        logger.info(a.as_dict())

    uhasdu()

    # logger.info(pivot_dict_object({"a": 2, "b": 2, "c": 3, "d": 4}, "id"))



__all__ = ["add_logging_handler_once"]

import logging
from logging import Logger

logger = logging.getLogger(__name__)


def add_logging_handler_once(logger: Logger, handler: object) -> bool:
    """A helper to add a handler to a logger, ensuring there are no duplicates.

    :param logger: Logger that should have a handler added.
    :type logger: logging.logger

    :param handler: Handler instance to be added. It will not be added if an
        instance of that Handler subclass already exists.
    :type handler: logging.Handler

    :returns: True if the logging handler was added, otherwise False.
    :rtype: bool
    """
    class_name = handler.__class__.__name__
    for handler in logger.handlers:
        if handler.__class__.__name__ == class_name:
            return False

    logger.addHandler(handler)
    return True



#!/usr/bin/env python3
import logging
from pathlib import Path
from typing import Union
from urllib.parse import urlparse

logger = logging.getLogger(__name__)
from warg.functions import sink

__all__ = ["get_requirements_from_file"]

try:
    # from packaging.requirements import Requirement

    # noinspection PyProtectedMember
    from pip._internal.network.session import PipSession

    # noinspection PyProtectedMember
    from pip._internal.req import parse_requirements

    # noinspection PyProtectedMember
    from pip._internal.req.req_file import ParsedRequirement

    # noinspection PyProtectedMember
    from pip._internal.utils.packaging import get_requirement

    def get_reqed(
        req,
        #: ParsedRequirement
    ):  # -> Requirement:
        """
        https://packaging.python.org/en/latest/specifications/direct-url/#example-pip-commands-and-their-effect-on-direct-url-json

        :param req:
        :type req: ParsedRequirement
        :return:
        :rtype: Requirement
        """

        req_ = req.requirement

        if req.is_editable:  # parse out egg=... fragment from VCS URL
            parsed = urlparse(req_)
            egg_name = parsed.fragment.partition("egg=")[-1]

            if not egg_name:
                egg_name = parsed.path.split("/")[-1]

            without_fragment = parsed._replace(fragment="").geturl()
            req_parsed = f"{egg_name} @ {without_fragment}"
        else:
            req_parsed = req_

        try:
            return get_requirement(req_parsed)
        except:
            return None

    def get_requirements_from_file(
        file_path: Union[str, Path], session: Union[str, PipSession] = "test"
    ):  # -> List[Requirement]
        """Turn requirements.txt into a list"""
        if isinstance(file_path, Path):
            file_path = str(file_path)

        parsed_reqs = [get_reqed(ir) for ir in parse_requirements(file_path, session=session)]

        return [p for p in parsed_reqs if p]

except Exception as e:  # (ModuleNotFoundError, ImportError) as e: #KeyError occurred
    logger.error(e)
    get_requirements_from_file = sink
    # logger.info('You version of python is to old!')

if __name__ == "__main__":
    logger.info(get_requirements_from_file(Path(__file__).parent.parent.parent / "requirements.txt"))



#!/usr/bin/env python3

__author__ = "Christian Heider Lindbjerg"
__doc__ = r"""

           Created on 09-09-2020
           """

__all__ = [
    "map_value_product",
    "map_product",
    "map_sorted",
    "map_permutations",
    "map_combinations",
    "map_combinations_with_replacement",
]

import itertools
import logging
from typing import Any, Generator, Mapping, Optional, Tuple

logger = logging.getLogger(__name__)


def map_value_product(mappings: Optional[Mapping]) -> Optional[Generator[dict, None, None]]:
    """

    :param mappings:
    :return:
    """
    if mappings is None:
        return None

    return (dict(zip(mappings, x)) for x in itertools.product(*mappings.values()))


def map_reversed(mapping: Optional[Mapping]) -> Optional[dict]:
    """

    :param mapping:
    :return:
    """
    if mapping is None:
        return None

    return dict(reversed(list(mapping.items())))


def map_sorted(mapping: Optional[Mapping], **kwargs) -> Optional[dict]:
    """

    :param mapping:
    :param kwargs:
    :return:
    """
    if mapping is None:
        return None

    return dict(sorted(mapping.items(), **kwargs))


def map_product(mapping: Mapping, repeat: int = 2) -> Any:
    """description"""
    yield from zip(
        itertools.product(mapping.keys(), repeat=repeat),
        itertools.product(mapping.values(), repeat=repeat),
    )


def map_permutations(mapping: Mapping, repeat: int = 2) -> Generator[Tuple[Any, ...], None, None]:
    """

    :param mapping:
    :param repeat:
    :return:
    """
    yield from zip(
        itertools.permutations(mapping.keys(), repeat),
        itertools.permutations(mapping.values(), repeat),
    )


def map_combinations(mapping: Mapping, repeat: int = 2) -> Generator[Tuple[Any, ...], None, None]:
    """

    :param mapping:
    :param repeat:
    :return:
    """
    yield from zip(
        itertools.combinations(mapping.keys(), repeat),
        itertools.combinations(mapping.values(), repeat),
    )


def map_combinations_with_replacement(
    mappings: Mapping, repeat: int = 2
) -> Generator[Tuple[Any, ...], None, None]:
    """

    :param mappings:
    :param repeat:
    :return:
    """
    if mappings is None:
        return None

    yield from zip(
        itertools.combinations_with_replacement(mappings.keys(), repeat),
        itertools.combinations_with_replacement(mappings.values(), repeat),
    )


if __name__ == "__main__":

    def asdijha() -> None:
        """
        :rtype: None
        """
        from warg import NOD

        a = NOD(a=[1], b=[4], c=[8])
        logger.info(f"ValueMapProduct{str(list(map_value_product(a.as_dict())))}")
        logger.info(f"MapProduct{str(list(map_product(a.as_dict())))}")
        logger.info(f"map_combinations{str(list(map_combinations(a.as_dict())))}")
        logger.info(f"map_permutations{str(list(map_permutations(a.as_dict())))}")
        logger.info(
            f"map_combinations_with_replacement{str(list(map_combinations_with_replacement(a.as_dict())))}"
        )

    def asdijhsadasdad() -> None:
        """
        :rtype: None
        """
        from warg import NOD

        a = NOD(a=[1, 2, 8], b=[4, 3, 99])
        logger.info(f"ValueMapProduct{str(list(map_value_product(a.as_dict())))}")
        logger.info(f"MapProduct{str(list(map_product(a.as_dict())))}")
        logger.info(f"map_combinations{str(list(map_combinations(a.as_dict())))}")
        logger.info(f"map_permutations{str(list(map_permutations(a.as_dict())))}")
        logger.info(
            f"map_combinations_with_replacement{str(list(map_combinations_with_replacement(a.as_dict())))}"
        )

    asdijha()
    # asdijhsadasdad()



#!/usr/bin/env python3

__author__ = "Christian Heider Lindbjerg"
__doc__ = r"""

           Created on 28-01-2021
           """

from warg import kws_sink, prod, sink

import logging

logger = logging.getLogger(__name__)


def test_a():
    logger.info(kws_sink("str"))
    logger.info(kws_sink(2))
    logger.info(kws_sink(2.2))

    logger.info(prod((2, 2)))

    logger.info(prod((2.2, 2.2)))

    logger.info(prod((2, 2.2)))

    logger.info(prod((2.2, 2)))

    logger.info(sink((2, 2), face=(2.2, 2)))



#!/usr/bin/env python3

__author__ = "Christian Heider Lindbjerg"
__doc__ = """
Generalised wrapper for delayed construction of class objects. Encapsulates kwargs and callable constructor with the option of modifying construction arguments before construction is finally performed.

"""

__all__ = ["GeneralisedDelayedKwargConstruction", "GDKC"]


import logging
from typing import Any, Callable, Mapping, MutableMapping, Sequence

logger = logging.getLogger(__name__)


class GeneralisedDelayedKwargConstruction:
    """
    A generalised class for setting up kwargs for later construction of an instance of an object
    [constructor, args, kwargs]
    """

    def __init__(self, constructor: Callable, *args: Any, **kwargs: Any):
        """
        [constructor, args, kwargs]

        :param constructor: The delayed callable, to be evaluated at __call__ or context __enter__
        :param args: arguments to use for evaluation, If only one is provided and is of typing mapping, it is assumed to be directly kwargs
        :param kwargs: arguments to use for evaluation
        """
        self.constructor: Callable = constructor
        assert len(args) < 2, f"Does not support multiple args, only a single mapping type"
        if len(args) == 1:
            assert isinstance(
                args[0], Mapping
            ), f"Arg[0] type is not a mapping type, was {type(args[0])} which is not supported"
        assert not (
            len(kwargs) > 0 and len(args) > 0
        ), f"Does not support both args and kwargs, both supplied args, {args} and {kwargs}"

        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], Mapping):
            self.kwargs: MutableMapping = args[0]
        elif len(kwargs) == 1 and next(iter(kwargs.keys())) == "kwargs":
            self.kwargs = kwargs["kwargs"]
        else:
            self.kwargs: MutableMapping = kwargs

    def __enter__(self) -> Any:
        self.instance = self.__call__()
        return self.instance.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb) -> Any:
        return self.instance.__exit__(exc_type, exc_val, exc_tb)

    def __call__(self, *args, **kwargs) -> Any:
        """

        Allows last minute override of kwargs

        :param args:
        :param kwargs:
        :return:
        """

        war = "".join(
            [
                f"Overwriting {k} with the value {v} in construction of {self.constructor}"
                for (k, v), b in zip(kwargs.items(), self.kwargs.keys())
                if k == b
            ]
        )
        if war != "":
            logger.warning(war)

        self.kwargs.update(kwargs)
        try:
            return self.constructor(*args, **self.kwargs)
        except TypeError as e:
            e.args += (f"in construction of {self.constructor}",)
            raise e


GDKC = GeneralisedDelayedKwargConstruction

if __name__ == "__main__":

    class UnreachableError(Exception):
        pass

    class A:
        """description"""

        def __init__(self, *args: Sequence[Any], **kwargs: MutableMapping[str, Any]):
            pass

    def stest_not_both() -> None:
        """
        :rtype: None
        """
        GeneralisedDelayedKwargConstruction(A, [1], a=2)

    def stest_kw() -> None:
        """
        :rtype: None
        """
        GeneralisedDelayedKwargConstruction(A, a=2)

    def stest_mapping() -> None:
        """
        :rtype: None
        """
        GeneralisedDelayedKwargConstruction(A, {"a": 2})

    def stest_mapping_and_args_fail() -> None:
        """
        :rtype: None
        """
        GeneralisedDelayedKwargConstruction(A, {"a": 2}, 1)

    def stest_mapping_and_args_fail_inv() -> None:
        """
        :rtype: None
        """
        GeneralisedDelayedKwargConstruction(A, 1, {"a": 2})

    stest_kw()
    stest_mapping()

    try:
        stest_mapping()  # Successful
        raise UnreachableError
    except UnreachableError:  # Is expected
        pass
    try:
        stest_mapping_and_args_fail()
        raise UnreachableError
    except AssertionError:
        pass
    try:
        stest_mapping_and_args_fail_inv()
        raise UnreachableError
    except AssertionError:
        pass
    try:
        stest_not_both()
        raise UnreachableError
    except AssertionError:
        pass



import logging
from typing import Any, Iterable, Union

logger = logging.getLogger(__name__)


def to_string_if_not_of_exact_type(gen: Iterable, type_: Iterable[type] = (int, float)) -> Union[str, Any]:
    """

    :param type_: Type for testing against
    :param gen: The iterable to be converted
    :return:
    """
    if not isinstance(type_, Iterable):
        type_ = [type_]

    for v in gen:
        if all([type(v) != t for t in type_]):
            yield str(v)
        else:
            yield v


def solve_type(d: Any) -> str:
    """
    Does not support size/length yet...

    :param d:
    :return:
    """
    if not isinstance(d, bool):
        if isinstance(d, int):
            return "integer"

        elif isinstance(d, float):
            return "double"

    return "string"


if __name__ == "__main__":
    logger.info(solve_type(True))

    logger.info(list(to_string_if_not_of_exact_type([True, 1, "A", 2.0])))



import importlib
import logging
import pkgutil
from pathlib import Path
from types import ModuleType
from typing import Any, Generator, Sequence

logger = logging.getLogger(__name__)


__all__ = ["iter_import_module", "get_submodules", "get_submodules_by_path"]


def iter_import_module(module_name: str, module_paths: Sequence[str]) -> Generator[ModuleType, Any, None]:
    logger.info(f"Found {module_name} in {module_paths}")

    yield from (
        importlib.import_module(f"{module_name}.{module_info.name}")
        for module_info in pkgutil.iter_modules(module_paths)
    )


def get_submodules(module: ModuleType) -> Generator[ModuleType, Any, None]:
    logger.info(f"Found {module.__path__} {module.__name__}")

    yield from iter_import_module(module.__name__, module.__path__)


def get_submodules_by_path(module_path: Path) -> Generator[ModuleType, Any, None]:
    assert module_path.exists(), f"{module_path} does not exist"

    yield from iter_import_module(module_path.stem, [str(module_path)])



#!/usr/bin/env python3

__author__ = "Christian Heider Lindbjerg"
__doc__ = r"""

           Created on 09/03/2020
           """

from warg import evaluate_context, kws_sink, prod, sink

import logging

logger = logging.getLogger(__name__)


def test_a():
    logger.info(evaluate_context(kws_sink, "str"))
    logger.info(evaluate_context(kws_sink, 2))
    logger.info(evaluate_context(kws_sink, 2.2))

    logger.info(evaluate_context(prod, (2, 2)))

    logger.info(evaluate_context(prod, (2.2, 2.2)))

    logger.info(evaluate_context(prod, (2, 2.2)))

    logger.info(evaluate_context(prod, (2.2, 2)))

    logger.info(evaluate_context(sink, (2, 2), face=(2.2, 2)))



#!/usr/bin/env python3

__author__ = "Christian Heider Lindbjerg"

__doc__ = """description"""
__all__ = ["to_british_english", "deamericanise"]

import logging
from types import MappingProxyType
from typing import Mapping

logger = logging.getLogger(__name__)
default_rules = MappingProxyType(
    {
        "ize": "ise",
        "yze": "yse",
        "iza": "isa",
        "aluminum": "aluminium",
        # 'se': 'ce',
        # 'og': 'ogue',
    }
)


def to_british_english(text: str, rules: Mapping = default_rules) -> str:
    """

    :param text:
    :type text: str
    :param rules:
    :type rules: Mapping
    :return: text
    :rtype: str
    """
    for r in rules.items():
        text = text.replace(*r)

    return text


def deamericanise(text: str) -> str:
    """
    Naively exchanges 'z' in english texts

    convert to 'British English'

    :param text: some text
    :type text: str
    :return: deamericanised text
    :rtype: str
    """
    return to_british_english(text)


if __name__ == "__main__":
    logger.info(
        deamericanise(
            "I analyzed websites in order to recognize the correct spelling of international organizations"
        )
    )



#!/usr/bin/env python3

__all__ = ["get_version"]

import datetime
import logging
from pathlib import Path

logger = logging.getLogger(__name__)


class NotGitException(Exception):
    pass


def get_version(version: str, append_time: bool = False, verbose: bool = False, context: Path = None) -> str:
    """

    :param context:
    :param verbose:
    :param version:
    :param append_time:
    :return:
    """

    if not version:
        version = "0.0.0"

    try:
        import subprocess

        if context is None:
            import inspect
            import os

            caller_parent = Path(os.path.abspath((inspect.stack()[1])[1])).resolve().parent

        else:
            context = Path(context)
            if context.is_file():
                caller_parent = context.parent
            else:
                caller_parent = context

        if (caller_parent / ".git").exists() and (caller_parent / ".git").is_dir():
            git_version = (
                subprocess.check_output(
                    [
                        "git",
                        "describe",
                        "--always",
                        # "--dirty",
                        "origin/HEAD",
                    ],
                    cwd=caller_parent,
                )
                .strip()
                .decode()
            )

            current_git_version = (
                subprocess.check_output(
                    [
                        "git",
                        "describe",
                        "--always",
                        "--dirty",
                    ],
                    cwd=caller_parent,
                )
                .strip()
                .decode()
            )

            if "dirty" in current_git_version:
                logger.warning(f"{caller_parent} git is dirty, {current_git_version}")

            if git_version.split("-")[0] != version:
                msg = f"{caller_parent} git version {git_version} does not match __version__" f" {version}"
                logger.warning(msg)
                assert git_version.split("-")[0] == version, msg

            else:
                if verbose:
                    msg = f"{caller_parent} git version {git_version} matches __version__" f" {version}"
                    logger.info(msg)

        else:
            raise NotGitException

    except:
        if append_time:
            try:
                now = datetime.datetime.now(datetime.UTC)
            except:
                now = datetime.datetime.utcnow()
            date_version = now.strftime("%Y%m%d%H%M%S")
            # date_version = time.time()

            if version:
                # Most git tags are prefixed with 'v' (example: v1.2.3) this is
                # never desirable for artefact repositories, so we strip the
                # leading 'v' if it's present.
                version = version[1:] if isinstance(version, str) and version.startswith("v") else version
            else:
                # The Default version is an ISO8601 compliant datetime. PyPI doesn't allow
                # the colon ':' character in its versions, and time is required to allow
                # for multiple publications to master in one day. This datetime string
                # uses the 'basic' ISO8601 format for both its date and time components
                # to avoid issues with the colon character (ISO requires that date and
                # time components of a date-time string must be uniformly basic or
                # extended, which is why the date component does not have dashes.)
                #
                # Publications using datetime versions should only be made from master
                # to represent the HEAD moving forward.
                logger.warning(
                    f"Environment variable VERSION is not set, only using datetime: {date_version}"
                )

                # warn(f'Environment variable VERSION is not set, only using timestamp: {version}')

            version = f"{version}.{date_version}"

    return version


if __name__ == "__main__":
    from warg import __version__

    logger.info(__version__)
    logger.info(get_version("1.2.7", verbose=True))



#!/usr/bin/env python3
__author__ = "Christian Heider Lindbjerg"
__doc__ = r"""
          Plugin

           Created on 13/06/2020
           """

import sys

if sys.version_info[:2] >= (3, 10):
    # pylint: disable=no-name-in-module
    # noinspection PyProtectedMember
    from importlib.metadata import entry_points, EntryPoint
else:
    # noinspection PyProtectedMember
    from importlib_metadata import entry_points, EntryPoint

from typing import Tuple, Generator, Any
import logging

logger = logging.getLogger(__name__)
__all__ = ["get_plugins", "get_static_plugins", "get_dynamic_plugins"]


def get_plugins(package_name: str) -> Tuple:
    """Returns a list specifying all known plugins.

    This includes both first-party, statically bundled plugins and
    dynamic plugins.

    Returns:
      The list of default first-party plugins.
    """
    return (*get_static_plugins(package_name), *get_dynamic_plugins(package_name))


def get_static_plugins(package_name: str) -> Tuple:
    """Returns a list specifying  default first-party plugins.

      DECLARE GLOBAL (PACKAGE_NAME)_PLUGINS tuple with entries

    Returns:
      The list of default first-party plugins.

    """
    v = f"{package_name.upper()}_PLUGINS"
    if v in globals():
        return globals()[v][:]
    return ()


def get_dynamic_plugins(
    package_name: str,
) -> Generator[EntryPoint, Any, None]:
    """Returns a list specifying dynamically loaded plugins.

    Returns:
      The list of dynamic plugins.

    [1]: https://packaging.python.org/specifications/entry-points/
    """

    return (
        entry_point.load() for entry_point in entry_points(group=f"{package_name}_plugins", name=package_name)
    )


if __name__ == "__main__":
    logger.info(get_plugins("warg"))

    logger.info(entry_points())



#!/usr/bin/env python3

import logging
from typing import Any, Callable, Generator, Iterable, List, Sequence, Sized, Tuple, Type

try:
    from itertools import pairwise
except:

    def pairwise(iterable: Iterable) -> Generator:
        # pairwise('ABCDEFG') → AB BC CD DE EF FG

        iterator = iter(iterable)
        a = next(iterator, None)

        for b in iterator:
            yield a, b
            a = b


logger = logging.getLogger(__name__)
__all__ = ["pairs", "chunks", "leaf_apply", "leaf_type_apply"]


def pairs(s: Iterable) -> Generator[Tuple[Any, Any], None, None]:
    """

    NOTE: Just use itertools.pairwise....

    Iterate over a list in overlapping pairs.

    Usage:
        lst = [4, 7, 11, 2]
        pairs(lst) yields (4, 7), (7, 11), (11, 2)

    https://stackoverflow.com/questions/1257413/1257446#1257446


    :param s: An iterable/list
    :return: Yields a pair of consecutive elements (lst[k], lst[k+1]) of lst. Last call yields (lst[-2], lst[-1]).
    """
    yield from pairwise(s)

    # i = iter(s)
    # prev = next(i)
    #
    # for item in i:
    #     yield prev, item
    #     prev = item


def leaf_apply(seq: Iterable, func: Callable) -> List:
    sub = []
    for element in seq:
        if isinstance(element, Iterable):
            sub.append(leaf_apply(element, func))
        else:
            sub.append(func(element))
    return sub


def leaf_type_apply(seq: Iterable, func: Callable, leaf_type: Type = Tuple) -> List:
    sub = []
    for element in seq:
        if not isinstance(element, leaf_type):
            sub.append(leaf_type_apply(element, func, leaf_type))
        else:
            sub.append(func(element))
    return sub


def chunks(lst: Sized, n: int) -> Any:
    """
    Yield successive n-sized chunks from lst.

     :param lst:
     :param n:
     :return:
    """
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


if __name__ == "__main__":
    logger.info(list(chunks(list(range(10)), 3)))
    logger.warning(list(pairs(list(range(10)))))



#!/usr/bin/env python3

__author__ = "heider"
__doc__ = r"""

           Created on 9/9/22
           """

__all__ = ["PrivateAttributeMixin"]

import logging
from typing import Any

logger = logging.getLogger(__name__)


class PrivateAttributeMixin:
    def __getitem__(self, key: str) -> Any:
        if key.startswith("_"):
            return None
        return getattr(self, key, None)



#!/usr/bin/env python3
import random
from typing import Callable, Iterable, List, Optional, Tuple

from warg import Number

__all__ = ["n_uint_mix", "n_uint_mix_generator_builder", "n_uint_mix_generator"]


import logging

logger = logging.getLogger(__name__)


def n_uint_mix(mix_max: Iterable[Number], mix_min: Optional[Iterable[Number]] = None) -> List[Number]:
    mix_max = list(mix_max)

    if mix_min is None:
        mix_min = [0] * len(mix_max)

    assert len(mix_min) == len(mix_max)

    for min_, max_ in zip(mix_min, mix_max):
        assert min_ <= max_

    return [random.randrange(min_, max_) for min_, max_ in zip(mix_min, mix_max)]


def n_uint_mix_generator(*mix, mix_min: Optional[Iterable[Number]] = None) -> Tuple[Number, ...]:
    if len(mix) == 1:
        if isinstance(mix, Iterable):
            mix = mix[0]

    while 1:
        yield n_uint_mix(mix, mix_min=mix_min)


def n_uint_mix_generator_builder(*mix: Number, mix_min: Optional[Iterable[Number]] = None) -> Callable:
    """Compatability code.."""

    if len(mix) == 1:
        if isinstance(mix, Iterable):
            mix = mix[0]

    def no_arg_generator() -> Tuple[Number, ...]:
        while 1:
            yield n_uint_mix(mix, mix_min=mix_min)

    return no_arg_generator


if __name__ == "__main__":
    logger.info([v for _, v in zip(range(9), iter(n_uint_mix_generator(255, 255)))])
    print([v for _, v in zip(range(9), iter(n_uint_mix_generator(255, 255, mix_min=(200, 200))))])



#!/usr/bin/env python3
from copy import deepcopy
from typing import Any, Generator, Iterable, Iterator

__author__ = "Christian Heider Lindbjerg"
__doc__ = r"""

           Created on 28/10/2019
           """

__all__ = ["unzip", "unzipper"]

import logging

logger = logging.getLogger(__name__)


def unzip(iterable: Iterable[Any]) -> Iterable[Any]:
    """description"""
    return zip(*iterable)


def unzipper(_iterable: Iterable[Iterable[Any]]) -> Iterable[Any]:
    """
    Unzips an iterable of an iterable

    Be careful, functionality maybe have undefined and unexpected behaviour

    :param _iterable:
    :return:Iterable
    """

    def check_next_iter(iterable_: Any) -> Any:
        """description"""
        if isinstance(iterable_, Iterable):
            try:
                a_ = next(iter(iterable_))
                if isinstance(a_, Iterable):
                    return a_
            except StopIteration:
                pass

    if isinstance(_iterable, Iterable):
        check_a = check_next_iter(check_next_iter(deepcopy(_iterable)))
        if check_next_iter(check_a):
            for a in _iterable:
                yield unzipper(a)
        elif check_a:
            for a in _iterable:
                yield unzip(a)
        else:
            yield from _iterable


if __name__ == "__main__":

    def recursive_eval(node: Any):
        """description"""
        if isinstance(node, (Iterable, Generator, Iterator)):
            gather = []
            for i in node:
                gather.append(recursive_eval(i))
            return gather
        return node

    def aasda() -> None:
        """
        :rtype: None
        """
        r = range(4)

        logger.info(0)

        a = [[[*r] for _ in r] for _ in r]
        logger.info(a)

        logger.info(1)

        for _, assd in zip(r, unzipper(a)):
            logger.info()
            logger.info(recursive_eval(assd))
            logger.info()

        for _, (a, *_) in zip(r, unzipper(a)):
            logger.info()
            logger.info(recursive_eval(a))
            logger.info()

        logger.info(2)

    def skad23() -> None:
        """
        :rtype: None
        """
        logger.info(0)
        zippy_once = zip(range(6), range(3))
        dsadsa = list(deepcopy(zippy_once))
        zippy_twice = zip(dsadsa, dsadsa)
        zippy_twice_copy = deepcopy(zippy_twice)
        asds = list(deepcopy(zippy_twice_copy))
        zippy_trice = zip(asds, asds)
        zippy_trice_copy = deepcopy(zippy_trice)

        logger.info(1)

        for aa in zippy_twice:
            logger.info(recursive_eval(aa))

        logger.info(2)

        for a1 in unzip(zippy_twice_copy):
            logger.info(recursive_eval(a1))

        logger.info(3)

        for a1 in unzip(zippy_once):
            logger.info(recursive_eval(a1))

        logger.info(4)

        for a1 in zippy_trice:
            logger.info(recursive_eval(a1))

        logger.info(5)

        for a1 in unzip(zippy_trice_copy):
            logger.info(recursive_eval(a1))

        logger.info(6)

    def skad() -> None:
        """
        :rtype: None
        """
        logger.info(0)
        zippy_once = zip(zip(range(6), range(3)))
        zippy_once_copy = deepcopy(zippy_once)
        dsadsa = list(deepcopy(zippy_once))
        zippy_twice = zip(dsadsa, dsadsa)
        zippy_twice_copy = deepcopy(zippy_twice)
        asds = list(deepcopy(zippy_twice_copy))
        zippy_trice = zip(asds, asds)
        zippy_trice_copy = deepcopy(zippy_trice)
        asds2323 = list(deepcopy(zippy_trice_copy))
        zippy_quad = zip(asds2323, asds2323)
        zippy_quad_copy = deepcopy(zippy_quad)

        logger.info(1)

        for aa in zippy_twice:
            logger.info(recursive_eval(aa))

        logger.info(2)

        for a1 in unzipper(zippy_twice_copy):
            logger.info(recursive_eval(a1))

        logger.info(3)

        for a1 in zippy_once_copy:
            logger.info(recursive_eval(a1))

        logger.info(4)

        for a1 in unzipper(zippy_once):
            logger.info(recursive_eval(a1))

        logger.info(5)

        for a1 in zippy_trice:
            logger.info(recursive_eval(a1))

        logger.info(6)

        for a1 in unzipper(zippy_trice_copy):
            logger.info(recursive_eval(a1))

        logger.info(7)

        for a1 in zippy_quad:
            logger.info(recursive_eval(a1))

        logger.info(8)

        for a1 in unzipper(zippy_quad_copy):
            logger.info(recursive_eval(a1))

        logger.info(9)

    aasda()

    logger.info()
    logger.info("asafasdw")
    logger.info()

    skad()
    # skad23()



#!/usr/bin/env python3
import json

try:
    from importlib.metadata import Distribution, PackageNotFoundError, PathDistribution
except (ModuleNotFoundError, ImportError) as e:
    from importlib_metadata import Distribution, PackageNotFoundError, PathDistribution

__all__ = [
    "dist_is_editable",
    "package_is_editable",
    "get_dist_package_location",
    "get_package_location",
]

from pathlib import Path
from typing import Optional
import logging

logger = logging.getLogger(__name__)
VERBOSE = False


# noinspection PyProtectedMember
def dist_is_editable(dist: Distribution) -> bool:
    """
    Return True if given Distribution is an editable installation.

    This function might still change alot

    """
    top_level = dist.read_text("top_level.txt")

    top_level_name = None
    if top_level:
        top_level_name = top_level.split("\n")[0].strip()

    else:  # assume top level namespace is the same as dist
        if isinstance(dist, PathDistribution):
            if hasattr(dist, "_normalized_name"):  # This is wacky...
                top_level_name = dist._normalized_name

        elif isinstance(dist, Distribution):
            if hasattr(dist, "name"):
                top_level_name = dist.name

    if top_level_name:
        if hasattr(dist, "_read_files_egginfo"):
            if dist._read_files_egginfo() is not None:
                if top_level_name == dist._path.parent.stem:
                    return True

    if hasattr(dist, "_read_files_distinfo"):
        if dist._read_files_distinfo() is not None:
            direct_url_str = dist.read_text("direct_url.json")
            if direct_url_str is not None:
                direct_url_json = json.loads(direct_url_str)
                if "dir_info" in direct_url_json:
                    if "editable" in direct_url_json["dir_info"]:
                        return direct_url_json["dir_info"]["editable"]

    return False


IGNORE = '''

def dist_is_editable(dist: Any) -> bool:
    """
    Return True if given Distribution is an editable installation."""
    for path_item in sys.path:
        egg_link = Path(path_item) / f"{dist.project_name}.egg-link"
        if egg_link.is_file():
            return True
    return False
'''


def package_is_editable(package_name: str) -> bool:
    """
    Return True if given Package is an editable installation.
    """
    try:
        dist = Distribution.from_name(package_name)

        return dist_is_editable(dist)

    except PackageNotFoundError as p:
        if VERBOSE:
            logger.info(p)


def get_package_location(package_name: str) -> Path:
    try:
        dist = Distribution.from_name(package_name)
        if dist:
            return get_dist_package_location(dist)

    except PackageNotFoundError as p:
        if VERBOSE:
            logger.info(p)


# noinspection PyProtectedMember
def get_dist_package_location(dist: Distribution) -> Optional[Path]:
    """
    FULL OF ASSUMPTIONS!

    :param dist:
    :return:
    """
    top_level = dist.read_text("top_level.txt")

    top_level_name = None
    if top_level:
        top_level_name = top_level.split("\n")[0].strip()

    else:  # assume top level namespace is the same as dist
        if isinstance(dist, PathDistribution):
            if hasattr(dist, "_normalized_name"):  # This is wacky...
                top_level_name = dist._normalized_name

        elif isinstance(dist, Distribution):
            if hasattr(dist, "name"):
                top_level_name = dist.name

    if top_level_name:
        if hasattr(dist, "_read_files_egginfo"):
            if dist._read_files_egginfo() is not None:
                if top_level_name == dist._path.parent.stem:
                    return dist._path.parent

    if hasattr(dist, "_read_files_distinfo"):
        if dist._read_files_distinfo() is not None:
            direct_url_str = dist.read_text("direct_url.json")
            if direct_url_str is not None:
                direct_url_json = json.loads(direct_url_str)
                if "dir_info" in direct_url_json:
                    if "editable" in direct_url_json["dir_info"]:
                        return Path(direct_url_json["url"])

    if top_level_name:
        package_location = dist._path.parent / top_level_name
        if package_location.exists() and package_location.is_dir():
            return package_location

    return None


if __name__ == "__main__":
    logger.info(package_is_editable(package_name="draugr"))
    logger.info(get_package_location(package_name="draugr"))

    logger.info(package_is_editable(package_name="warg"))
    logger.info(get_package_location(package_name="warg"))

    logger.info(package_is_editable(package_name="apppath"))
    logger.info(get_package_location(package_name="apppath"))

    logger.info(get_package_location(package_name="numpy"))

    logger.info(package_is_editable(package_name="Pillow"))
    logger.info(get_package_location(package_name="Pillow"))
    logger.info(get_package_location(package_name="pillow"))



#!/usr/bin/env python3

__author__ = "Christian Heider Lindbjerg"
__doc__ = r"""

           Created on 14/01/2020
           """

__all__ = [
    "kws_sink",
    "sink",
    "prod",
    "collate_first_dim",
    "call_identity",
    "args_sink",
    "identity",
    "invert_shallow_mapping",
    "flip_two_level_mapping",
    "swap_mapping_order",
    "nop",
    "empty_str",
    "list_keys",
    "first_key",
    "last_key",
    "to_list",
    "to_tuple",
    "recurse_replace_empty",
    "text_in_file",
    "int_limits",
    "flatten_mapping",
    "most_common_substrings",
    "mappings_agreement_reduce",
]

import ctypes
import logging
import operator
import sys
import webbrowser
from collections import defaultdict
from copy import deepcopy
from difflib import SequenceMatcher
from functools import reduce
from itertools import product
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Iterator, List, Mapping, Optional, Sequence, Tuple, Union

import requests

from warg.contexts import Suppress
from warg.decorators import drop_unused_kws
from warg.typing_extension import Number

logger = logging.getLogger(__name__)


def int_limits(c_int_type: Any) -> Tuple[int, int]:
    signed = c_int_type(-1).value < c_int_type(0).value
    bit_size = ctypes.sizeof(c_int_type) * 8
    signed_limit = 2 ** (bit_size - 1)
    return (-signed_limit, signed_limit - 1) if signed else (0, 2 * signed_limit - 1)


TO_LIST_MAX_RECURSION_LIMIT = int_limits(ctypes.c_int)[-1]


def recurse_replace_empty(iterable: Iterable) -> Optional[Iterable]:
    if isinstance(iterable, str) or not isinstance(iterable, Iterable):
        return iterable

    if iterable:
        if isinstance(iterable, Mapping):
            return {k: recurse_replace_empty(v) for k, v in iterable.items()}

        # noinspection PyArgumentList
        return type(iterable)(
            recurse_replace_empty(i) for i in iterable
        )  # contstruct type of iterable with replaced empties

    return None


def list_keys(d: Dict) -> List[Any]:
    return list(d.keys())


def first_key(d: Dict) -> Any:
    return list_keys(d)[0]


def last_key(d: Dict) -> Any:
    return list_keys(d)[-1]


def empty_str() -> str:
    """

    :return:
    :rtype:
    """
    return ""


def nop() -> None:
    """
    :rtype: None
    """
    pass


def identity(a: Any) -> Any:
    """description"""
    return a


@drop_unused_kws
def kws_sink(*args) -> Tuple[Any, ...]:
    """
    Returns args without any modification what so ever. Drops kws
    :return:"""
    return args


def call_identity(*args, **kwargs) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
    """

    :param args:
    :param kwargs:
    :return:"""
    return args, kwargs


# noinspection PyUnusedLocal
def args_sink(*args, **kwargs) -> Dict[str, Any]:
    """

    :param args:
    :param kwargs:
    :return:"""
    return kwargs


# noinspection PyUnusedLocal
def sink(*args, **kwargs) -> None:
    """
    Returns None, but accepts everything

    :param args:
    :param kwargs:
    :return:"""
    return


def prod(iterable: Iterable[Number]) -> Number:
    """
    Calculate the product of an Iterable, of int or floats
    :param iterable:
    :return:"""
    return reduce(operator.mul, iterable, 1)


def collate_first_dim(batch: Iterable) -> Tuple:
    """

    :param batch:
    :return:"""
    return tuple(zip(*batch))


def invert_shallow_mapping(m: Mapping) -> Dict:
    """

    :param m:
    :return:
    """
    return {v: k for k, v in m.items()}


def flip_two_level_mapping(m: Mapping) -> Dict:
    """
    result = {}
    [result.setdefault(a, {}).update({k:b}) for k, v in m.items() for a, b in v.items()]
    return result

    :param m:
    :return:
    """

    flipped = defaultdict(dict)
    for key, val in m.items():
        for sub_key, sub_val in val.items():
            flipped[sub_key][key] = sub_val
    return flipped


def swap_mapping_order(m: Mapping, order: Sequence[int]) -> Mapping:
    """

    :param m:
    :param order:
    :return:
    """
    order = [*order]

    def deep_swap(dict_, level):
        """

        :param dict_:
        :param level:
        :return:
        """

        def swap_two_level_dict(a):
            """

            :param a:
            :return:
            """
            b = defaultdict(dict)
            for key1, value1 in a.items():
                for key2, value2 in value1.items():
                    b[key2].update({key1: value2})
            return b

        dict_ = deepcopy(dict_)
        if level == 0:
            dict_ = swap_two_level_dict(dict_)
        else:
            for key in dict_:
                dict_[key] = deep_swap(dict_[key], level - 1)
        return dict_

    for pas_no in range(len(order) - 1, 0, -1):
        for i in range(pas_no):
            if order[i] > order[i + 1]:
                temp = order[i]
                order[i] = order[i + 1]
                order[i + 1] = temp
                m = deep_swap(m, i)
    return m


def chain_filter(it: Iterable, *filters: Callable) -> Iterator:
    """
    Apply a sequence of callables to an iterable through filter; filtering the iterable to the subset of a callable
    returns

    Args:
        it (Iterable):
            iterable to be filtered
        filters (Callable):
            The filter callables

    Returns:
        Iterator:
            returns an iterator yielding those items of iterable for which all(filters(item)) is true. If filters are
            None, return the items that are true.
    """
    for f in filters:
        it = filter(f, it)
    return it


def chain_apply(it: Iterable, *callables: Callable) -> Iterable[Any]:
    """
    Apply a sequence of callables to an iterable; apply the iterable sequentially in callables order

    Args:
        it (Iterable):
            iterable to be applied to
        callables (Callable):
             The applying callables

    Returns:
        Iterable:
            returns the iterable with all the callables applied.
    """
    for f in callables:
        it = f(it)
    return it


def to_list(x: Union[Iterable, Any]) -> Union[List, Any]:
    if False:
        if x is None:
            return []

    if TO_LIST_MAX_RECURSION_LIMIT:
        original = sys.getrecursionlimit()
        sys.setrecursionlimit(TO_LIST_MAX_RECURSION_LIMIT)

    if not isinstance(x, Iterable):
        return x

    i = iter(x)

    try:
        val = next(i)
    except StopIteration:
        return []
    finally:
        if TO_LIST_MAX_RECURSION_LIMIT:
            sys.setrecursionlimit(original)

    return [to_list(val)] + to_list(i)


def to_tuple(x: Union[Iterable, Any]) -> Union[Tuple, Any]:
    if False:
        if x is None:
            return ()

    if TO_LIST_MAX_RECURSION_LIMIT:
        original = sys.getrecursionlimit()
        sys.setrecursionlimit(TO_LIST_MAX_RECURSION_LIMIT)

    if not isinstance(x, Iterable):
        return x

    i = iter(x)

    try:
        val = next(i)
    except StopIteration:
        return ()
    finally:
        if TO_LIST_MAX_RECURSION_LIMIT:
            sys.setrecursionlimit(original)

    return (to_tuple(val), *to_tuple(i))


@Suppress(FileNotFoundError)
def text_in_file(text: str, filename: Path) -> bool:
    if filename.exists():
        return any(text in line for line in filename.open())

    return False


def flatten_mapping(mapping: Mapping[str, Any], seperator: str = "_") -> Mapping[str, Any]:
    """
    iterates and recursively flattens nested mappings by appending keys for each level


    :param seperator:
    :param mapping: nested mapping to flatten
    :return: a single level mapping with appended level keys
    """
    out_dict = {}
    for k, v in mapping.items():
        if isinstance(v, Mapping):
            out_dict.update(**{f"{k}{seperator}{ki}": vi for ki, vi in flatten_mapping(v, seperator).items()})
        else:
            out_dict[k] = v

    return out_dict


def most_common_substrings(
    strings: Iterable[str], min_common_substring: int = 2, min_common_count: int = 2
) -> List[str]:
    """
    Finds common substrings in Iterable of strings, ordered in most to least occurring

    :param strings:
    :param min_common_substring:
    :param min_common_count:
    :return:
    """
    name_matches = []

    for z1, z2 in product(strings, strings):
        if z1 != z2:
            match = SequenceMatcher(isjunk=None, a=z1, b=z2).find_longest_match(
                alo=0, ahi=len(z1), blo=0, bhi=len(z2)
            )
            match_string = z1[match.a : match.a + match.size]
            if len(match_string) >= min_common_substring:
                name_matches.append(match_string)

    return sorted({i for i in name_matches if name_matches.count(i) >= min_common_count})


def mappings_agreement_reduce(m1: Mapping[Any, Any], m2: Mapping[Any, Any]) -> Mapping[Any, Any]:
    """
    takes two mappings and reduces it to one mapping, if the agree or one of them is None take one or the other. if
    they disagree the resulting value is None.

    :param m1: A mapping
    :param m2: Another mapping
    :return: a single mapping
    """
    out = {}
    for k, v in m1.items():
        if k not in m2 or v == m2[k] or not m2[k]:
            out[k] = v
        elif not v and k in m2:
            out[k] = m2[k]
        else:
            out[k] = None

    for k, v in m2.items():
        if k not in out:
            out[k] = v

    return out


def open_uri_resource(uri: str) -> str:
    """Opens the default web browser.
    Qt offers PyQt5.QtWebEngineWidgets (QWebEngineView, QWebEngineSettings) but they are not
    available from pyQGIS

    """

    path = Path(uri)

    if path.exists() and path.is_file():
        uri = f"file:///{path.absolute()}"
        msg = uri
        webbrowser.open_new_tab(uri)
    else:
        try:
            uri = f'https://{uri.replace("https://","")}'
            r = requests.head(uri)
            if r.ok:  # it is a boolean
                try:
                    webbrowser.open_new_tab(uri)
                    msg = uri
                except webbrowser.Error as e:
                    msg = f"Webbrowser error: {e}"

            elif r.status_code == 404:
                msg = f'HTTP Error 404: Page not found.<br>The following URL may be broken or dead:<br><br><a href="{uri}">{uri}</a>'

            else:
                msg = f'Error with URL<br><br><a href="{uri}">{uri}</a>'

        except requests.ConnectionError as e:
            msg = f'URL <a href="{uri}">{uri}</a> could not be opened.<br><br>Is your internet connection up and working?'

    return msg


if __name__ == "__main__":

    def asud() -> None:
        """
        :rtype: None
        """
        a = {"b": 1, "h": 2}
        logger.info(invert_shallow_mapping(a))

    def asjdnasid() -> None:
        """
        :rtype: None
        """
        a = {
            "b": {"c": {"d": [0, 1, 2], "e": [3, 4, 5, 6]}, "f": [7, 8], "g": [9]},
            "h": {"j": [10, 11]},
        }
        logger.info(flip_two_level_mapping(a))

    def asidj() -> None:
        """
        :rtype: None
        """
        test_dict = {
            "a": {"c": {"e": 0, "f": 1}, "d": {"e": 2, "f": 3}},
            "b": {"c": {"g": 4, "h": 5}, "d": {"j": 6, "k": 7}},
        }
        result = swap_mapping_order(test_dict, [2, 0, 1])
        logger.info(result)

    def i8jsadij():
        ij = ([10, 29], [(2, 3, 4), [[12, 4, 5]], ((2, 92, 90))], [])
        logger.info(to_list(ij))

    def i8jsadi2j():
        ij = ([10, 29], [(2, 3, 4), [[12, 4, 5]], ((2, 92, 90))], [])
        logger.info(to_tuple(ij))

    # i8jsadij()
    # i8jsadi2j()
    # asud()
    # asidj()
    # asjdnasid()

    # logger.info(open_uri_resource(__file__))
    logger.info(open_uri_resource("dr.dk"))



__all__ = ["unload_modules"]

import importlib
import sys

BUILT_IN = [*sys.builtin_module_names, *sys.stdlib_module_names]


def unload_modules(*modules: str) -> None:
    if not modules:
        modules = sys.modules.copy()

    for sm in modules.keys():
        if sm in BUILT_IN:
            continue

        del sys.modules[sm]

        for mod in modules.values():
            try:
                delattr(mod, sm)
            except AttributeError:
                pass

    importlib.invalidate_caches()
    # sys._clear_internal_caches()


if __name__ == "__main__":

    def main():
        import argparse

        print(argparse)

        unload_modules()

        print(sys.modules)

    main()



#!/usr/bin/env python3

__author__ = "heider"
__doc__ = r"""

           Created on 12/19/22
           """

__all__ = [
    "install_requirements_from_file",
]

import logging

import os
import subprocess
import sys
from enum import Enum, auto
from pathlib import Path
from typing import Optional

logger = logging.getLogger(__name__)

# from warg import is_windows # avoid dependency import not standard python pkgs.
CUR_OS = sys.platform
IS_WIN = any(CUR_OS.startswith(i) for i in ["win32", "cygwin"])
IS_MAC = CUR_OS.startswith("darwin")
VERBOSE = False


# @passes_kws_to(subprocess.check_call)
def catching_callable(*args, **kwargs) -> None:
    try:
        # subprocess.check_call(*args, **kwargs)
        output = subprocess.check_output(*args, **kwargs)
        # subprocess.run(*args,**kwargs)
    except subprocess.CalledProcessError as e:
        output = (e.stderr, e.stdout, e)
    logger.warning(output)


SP_CALLABLE = catching_callable  # subprocess.call
DEFAULT_PIP_INDEX = os.environ.get("PIP_INDEX_URL", "https://pypi.org/pypi/")


class InstallStrategyEnum(Enum):
    """ """

    uninstall_first = auto()  # TODO: HOW MANY TIME TO DO THIS?
    just_install = auto()


def is_pip_installed() -> bool:
    pip_present = True
    try:
        import pip
    except ImportError:
        pip_present = False
    return pip_present


def get_embedded_python_interpreter_path() -> Optional[Path]:
    """

    :return: The path of the qgis python interpreter
    :rtype: Optional[Path]
    """
    interpreter_path = Path(sys.executable)
    fallback = True

    if IS_WIN:
        try_path = interpreter_path.parent / "python.exe"
        if not try_path.exists():
            try_path = interpreter_path.parent / "python3.exe"
            if not try_path.exists():
                logger.error(f"Could not find python {try_path}")
                if not fallback:
                    return None
            else:
                return try_path
        else:
            return try_path

    elif IS_MAC:
        try_path = interpreter_path.parent / "bin" / "python"
        if not try_path.exists():
            try_path = interpreter_path.parent / "bin" / "python3"
            if not try_path.exists():
                logger.error(f"Could not find python {try_path}")
                if not fallback:
                    return None
            else:
                return try_path
        else:
            return try_path

    return interpreter_path


def install_pip_if_not_present(
    always_upgrade: bool = True, install_strategy: InstallStrategyEnum = InstallStrategyEnum.just_install
) -> None:
    if install_strategy == InstallStrategyEnum.uninstall_first:
        ...  # TODO: Implement

    if not is_pip_installed() or always_upgrade:
        if False:
            import ensurepip

            ensurepip.bootstrap(upgrade=True)
        else:
            SP_CALLABLE(
                [
                    str(get_embedded_python_interpreter_path()),
                    "-m",
                    "ensurepip",
                    "--upgrade",
                ]
            )


class UpgradeStrategyEnum(Enum):
    """
    eager - all packages will be upgraded to the latest possible version. It should be noted here that pip’s current
    resolution algorithm isn’t even aware of packages other than those specified on the command line, and those
    identified as dependencies. This may or may not be true of the new resolver.

    only-if-needed - packages are only upgraded if they are named in the pip command or a requirement file (i.e,
    they are direct requirements), or an upgraded parent needs a later version of the dependency than is currently
    installed.

    to-satisfy-only (undocumented, please avoid) - packages are not upgraded (not even direct requirements) unless the
    currently installed version fails to satisfy a requirement (either explicitly specified or a dependency).

    This is actually the “default” upgrade strategy when --upgrade is not set, i.e. pip install AlreadyInstalled and pip
    install --upgrade --upgrade-strategy=to-satisfy-only AlreadyInstalled yield the same behavior.
    """

    eager = "eager"
    to_satisfy_only = "to-satisfy-only"
    only_if_needed = "only-if-needed"


# subprocess.Popen(**ADDITIONAL_PIPE_KWS)
# ADDITIONAL_PIPE_KWS = dict(stderr=subprocess.PIPE,stdout=subprocess.PIPE, stdin=subprocess.PIPE)


def pip_programmatic_install2(package: str) -> None:
    """
    not supported
    :param package:
    :return:
    """
    import importlib

    try:
        importlib.import_module(package)
    except ImportError:
        import pip

        if hasattr(pip, "main"):
            pip.main(["install", package])
        else:
            pip._internal.main(["install", package])
    finally:
        globals()[package] = importlib.import_module(package)
        import site
        from importlib import reload

        reload(site)


def install_requirements_from_file(
    requirements_path: Path,
    upgrade: Optional[bool] = None,
    upgrade_strategy: UpgradeStrategyEnum = UpgradeStrategyEnum.only_if_needed,
) -> None:
    """
    Install requirements from a requirements.txt file.

    :param upgrade:
    :param upgrade_strategy:
    :param requirements_path: Path to requirements.txt file.
    :rtype: None
    """
    requirements_file_parent_directory = str(requirements_path.parent.as_posix())

    os.environ["REQUIREMENTS_FILE_PARENT_DIRECTORY"] = requirements_file_parent_directory

    req_path_str = str(requirements_path)

    args = ["install", "-r", req_path_str]

    if True:  # No progress bar
        args += ["--progress-bar", "off"]

    if True:
        args += ["--user"]

    if upgrade:
        args += ["--upgrade"]

    if True:
        args += ["--ignore-installed"]

    if upgrade_strategy:
        args += ["--upgrade-strategy", upgrade_strategy.value]

    if False:
        import pip

        pip.main(args)

    elif False:
        SP_CALLABLE(["pip"] + args)

    elif True:
        install_pip_if_not_present()

        if is_pip_installed():
            SP_CALLABLE([str(get_embedded_python_interpreter_path()), "-m", "pip", *args])

        else:
            logger.info("PIP IS STILL MISSING!")


if __name__ == "__main__":
    logger.info(get_embedded_python_interpreter_path())



#!/usr/bin/env python3

__author__ = "Christian Heider Lindbjerg"
__doc__ = r"""

           Created on 30/03/2020
           """

from pathlib import Path

with open(Path(__file__).parent / "README.md") as this_init_file:
    __doc__ += this_init_file.read()
# del Path

from .look_up_table import *
from .property_caching import *



#!/usr/bin/env python3

__author__ = "Christian Heider Lindbjerg"
__doc__ = r"""

           Created on 23/07/2020
           """

__all__ = ["recursive_flatten"]


import logging
from typing import Iterable, Sequence

logger = logging.getLogger(__name__)


def recursive_flatten_seq(seq: Sequence) -> Sequence:
    """Depth first flatten"""
    if not seq:  # is empty Sequence
        return seq
    if isinstance(seq[0], Sequence):
        return (*recursive_flatten(seq[0]), *recursive_flatten(seq[1:]))
    return (*seq[:1], *recursive_flatten(seq[1:]))


def recursive_flatten(sequence: Iterable) -> Iterable:
    """
    Depth first flattens iterable

    >>> list(recursive_flatten([1, [2], 3]))
    [1, 2, 3]
    >>> list(recursive_flatten([1, [2], [3, [4]]]))
    [1, 2, 3, 4]
    >>> list(recursive_flatten((([[None]], 2), (2,), 2)))
    [None, 2, 2, 2]
    """
    for element in sequence:
        if isinstance(element, Iterable) and not isinstance(element, str):
            yield from recursive_flatten(element)
        else:
            yield element


if __name__ == "__main__":
    logger.info(list(recursive_flatten((((2,), 2), (2,), 2))))
    logger.info(list(recursive_flatten(((("2",), 2), (2,), 2))))
    logger.info(list(recursive_flatten((([[None]], 2), (2,), 2))))

    logger.info(list(recursive_flatten((([[None]], 2), (2,), 2))))
